/* ***** BEGIN LICENSE BLOCK *****
 * Licensed under Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * Full Terms at http://mozile.mozdev.org/license.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Playsophy code.
 *
 * The Initial Developer of the Original Code is Playsophy
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *		Karl Guertin <grayrest@grayrest.com>
 *		James A. Overton <james@overton.ca>
 *		Christian Stocker (Bitflux)
 *		Paul Everitt
 *		David Palm
 *
 * ***** END LICENSE BLOCK ***** */

/**
@fileoverview
This file is responsible for handling the keyboard events.

<p>NB: Firefox 1.0.6 (or thereabouts). You can no longer create synthetic keypress
events using .createEvent("KeyEvents"), initKeyEvent, dispatchEvent.

<p>To Do:
<p>proper handling of navigation via eDOM.
<p>command key overrides - enter, tab, etc
*/


/********* Handler *********/
/**
Allow Mozile to intercept all keypress events for the specified <code>target</code> document.
Uses event capture. 
*/
function mozileAddKeypressListener(target) {
target.document.addEventListener("keypress", mozileKeys, true);
}
/**
Stop Mozile from intercepting keypress events.
*/
function mozileRemoveKeypressListener(target) {
target.document.removeEventListener("keypress", mozileKeys, true);
}

/**
Keyboard object.
Keyboard handling in the eDOM is currently incomplete.
A timeout function is included to update the toolbar after browser keypresses
that aren't handled by the eDOM. These make editing easier if you wish to
edit large blocks of text. For setting up an editing situation to edit
individual words, for example, using the eDOM navigation to constrain the
caret to the editable region would probably be better. 

@constructor
*/
function MozileKeyboard() {
	this.timer=null;
	this.refresh=false;
}
MozileKeyboard.prototype.startRefresh = function () {
	this.refresh = true;
	this.timer = setInterval(mozileKeyboardRefresh, 10);
}
MozileKeyboard.prototype.clearRefresh = function () {
	if (this.timer)	clearInterval(this.timer);
	this.refresh = false;
}
//If selection hasn't changed, need to wait longer?
MozileKeyboard.prototype.doRefresh = function () {
	if (this.refresh) {
		this.clearRefresh();
		mozileDetect();
	}
}

/**
Must have Interval Timeout function as a global function? Couldn't get it to
work with the function as a method of mozileKeyboard.
*/
function mozileKeyboardRefresh() {
	mozileKeyboard.doRefresh();
}

var mozileKeyboard = new MozileKeyboard();


/**
Key press event listener.
Handles keyboard events.
*/
function mozileKeys(event) {
	if (!mozilePrefs.getBoolPref("mozile.onBool")) {return;}
	var t=event.target;

 try {
	if (mozileDetectOverride(t)) {
		return;
	}
	if(mozileDetect()) {
		mozileDetectToolbar(); //Show toolbar and turn caret on.
	}	else {
		mozileNoDetectToolbar( t ); //hide the toolbar (if autohide is on) and restore the caret state.
		return; //ignore keypress.
	}

	//Mozile detected...

	if ( !(("supportsEDOM" in mozileTarget.document) && mozileTarget.document.supportsEDOM()) )
		return; //Wait if Javascript loaded but not yet ready for action. 

	var handled;
	// Ctrl OR Meta key (Meta is Mac OSX standard)
	if(event.ctrlKey || event.metaKey) {
		handled = ctrlKeyPressHandler(event);
	}else if (event.altKey) {
		handled = altKeyPressHandler(event);
	}else{ // All other keys...
		handled = nonctrlKeyPressHandler(event);
	} 

	// If Mozile has handled the event try and make sure nothing else does.
	if(handled){
		event.stopPropagation(); //This IS needed.
		event.preventDefault();
//		event.returnValue = false;
	}
 }catch(e){
  alert("Keyboard exception trap: "+e);
 }
}

/**
Ctrl+Key Commands.
Handles control+key commands if the override preference is set.
Note: When the control key is pressed the Caps Lock button is ignored.

<ul>
<li>Toolbar defaults
	<ul>
	<li>Ctrl-x : Cut
	<li>Ctrl-v : Paste
	<li>Ctrl-z : Undo
	<li>Ctrl-shift-z : Redo
	<li>Ctrl-shift-s : Show save dialog.
	</ul>
<li>Ctrl-s : Save using current settings.
<li>MES specified toolbarbutton keys
</ul>

<p>Should only update (toolbar) etc if required.
<p>TO DO: Note that Ctrl+LeftArrow and Ctrl+RightArrow could move cursor out of the
editable area or into a different tag. MozileDetect should be called AFTER these events.
Ctrl+Up/Down don't seem to do anything.
Ctrl+PageUp/Down/Tab switch window. Lose caret when return to Mozile and need to restore it.

Pressing Ctrl+shift+key doesn't always generate a keypress event for some reason. 

@return true if key press handled, false if not a recognised command.
@type boolean
*/
function ctrlKeyPressHandler(event)
{
	if(!mozilePrefs.getBoolPref("mozile.acceleratorBool")){
		return false;
	}else{
		if(event.charCode) {
			var result=true;
			var update=true;
			var C = String.fromCharCode(event.charCode);
			if(C=="s"){ 
				mozileSave();
			}else if(!mozileCtrlCommand(C)){
				update=false;
			}
			if(update) {
				mozileDetect();
				mozile.source.browserInput(mozileTarget, mozileSelection);
			}
		}else if(event.keyCode){
			var key=event.keyCode;
			if((key==37 || key==39)&& event.shiftKey)
				result=false;	//Ignore Ctrl+shift+left arrow/right arrow.
		}
		return result;
	}
}

/**
Scan Mozile toolbar buttons for the requested keyboard shortcut and execute
the command if found. The keyboard shortcuts are specified in the MES file.
Only the first button found with the shortcut will be executed.
i.e. It is up to you to ensure you don't put duplicates in the MES.

@param {String} C : charCode for the key pressed.
@return true if command executed, otherwise false.
@type boolean
*/
function mozileCtrlCommand( C ) {
	var toolbar = document.getElementById('mozile-toolbar');
	var b = toolbar.getElementsByTagName('toolbarbutton');
	var i,k;
	for (i=0; i<b.length; i++) {
		if (b[i].hasAttribute('key')) {
			k = b[i].getAttribute('key');
			if ( k.indexOf(C) != -1  ) {
				b[i].doCommand();
				return true;
			}
		}
	}
	b = toolbar.getElementsByTagName('menuitem');
	for ( i=0; i<b.length; i++) {
		if (b[i].hasAttribute('key')) {
			k = b[i].getAttribute('key');
			if ( k.indexOf(C) != -1  ) {
				b[i].doCommand();
				return true;
			}
		}
	}
	return false;
}

/**
Alt+key commands.

<p>Note that toolbar toolbarbuttons have the accesskey property but the feature
does not seem to be implemented in the browser.

@return true if key press handled, false if not a recognised command.
@type boolean
*/
function altKeyPressHandler(event)
{
	if(!mozilePrefs.getBoolPref("mozile.acceleratorBool")){
		return false;
	}else{
		if(event.charCode) {
			var C = String.fromCharCode(event.charCode);
			var c = C.toLowerCase();
			var pop,menu;

			if(c==mozileTools.menuKey){
				pop=document.getElementById("mozile-menu-popup");
				menu=document.getElementById("mozile-menu");
				pop.showPopup( menu, -1, -1, "popup","bottomleft","topleft" );
			}
			else if(c==mozileTools.commandMenuKey){
				pop=document.getElementById("mozile-command-menu-popup");
				if (pop) {
					menu=document.getElementById("mozile-command-menu");
					pop.showPopup( menu, -1, -1, "popup","bottomleft","topleft" );
				}
			}
		}
		return true;
	}
}

/**
Everything except Ctrl+? commands.

<p>Note: Mozile needs to control left,right,up and down key handling to ensure that
the cursor remains in the editable area. Also, the browser does not handle left/right
traversal through elements such as <img/> correctly.

<p>Mozile up/down handling can not yet handle lines correctly but moves to the same offset
in the next block. Hence, you may prefer to use the browser up/down keys and comment out
the Mozile handling code. The lines to comment out are  indicated with "###".
Note, however, that in some circumstances this can cause a detection error and eDOM to
throw an exception.

<p>Home and End: with eDOM go to the start/end of a block rather than the start/end of the displayed line.
With the default browser handling Home goes to the start of the page!
Shift-Home/End however work correctly to select to the start/end of the line.

<p>Page Up/Down can move you out of the editable area. 

<p>Non-break space: Should be able to enter non-break space as a valid character.
This normally works but in some cases there is a conflict with the use of nbs as the editing token.
Also note that the keyboard event for the key combination ctrl+shift+spacebar
is not recognised. Hence just shift+spacebar is used instead to insert a nbsp. 

<p>POST04:
<ul>
<li>carefully move selectEditableRange in here
<li>distinguish editable range of deleteOne at start of line and deleteOne
 on same line [need to stop merge but allow character deletion]. Perhaps
 need to change eDOM granularity.

@return true if Mozile recognizes the key press and attempts to handle it.
@type boolean
*/
function nonctrlKeyPressHandler(event)
{
	var result=true;
	var update=true;
	var key = event.keyCode;
	var shift = event.shiftKey;
	var C=event.charCode;
	var pre;

	if(key){
		if(key==33 || key==34 || key==35 || key==36 || key==38 || key==40 || (shift && key>=33 && key<=40) ) { //+page up=33 + page down=34
			mozileKeyboard.startRefresh();		
		}
		if(key==37 && !shift){ //(DOM_VK_LEFT)
			pre = mozileSelection.moveInsertionPoint(mozileTarget.Selection.MOVE_BACK_ONE);
			if ( pre===null ) {
				result=update=false;
			}
		}else if(key==39 && !shift){ //(DOM_VK_RIGHT)
			pre = mozileSelection.moveInsertionPoint(mozileTarget.Selection.MOVE_FORWARD_ONE);
			if ( pre===null ) {
				result=update=false;
			}
//		}else if(key==38 && !shift) { //(DOM_VK_UP)                                 //###
//			mozileSelection.moveInsertionPoint(mozileTarget.Selection.MOVE_UP_ONE);   //###
//		}else if(key==40 && !shift) { //(DOM_VK_DOWN)                               //###
//			mozileSelection.moveInsertionPoint(mozileTarget.Selection.MOVE_DOWN_ONE); //###
//		}else if(key==36 && !shift){ //(DOM_VK_HOME)
//			mozileSelection.moveInsertionPoint(mozileTarget.Selection.MOVE_START_LINE);
//		}else if(key==35 && !shift) { //(DOM_VK_END)
//			mozileSelection.moveInsertionPoint(mozileTarget.Selection.MOVE_END_LINE);
		}else if(key==13 || key==14){	//(DOM_VK_RETURN, DOM_VK_ENTER)
			mozileDo("Return key");
			//IMPORTANT: Switch off caret to avoid cursor flicker as nodes are modified.
			mozilePrefs.setBoolPref('accessibility.browsewithcaret', false); 
			mozileSelection.splitXHTMLLine(shift); 
//			mozileSelection.splitXHTMLLine(false);	/*Use this line if you want to ignore shift+return.*/
		}else if(key==9){	//(DOM_VK_TAB)
			mozileDoTyping();
			mozilePrefs.setBoolPref('accessibility.browsewithcaret', false);
			mozileSelection.insertTab(shift);
		}else if(key==8){	// (DOM_VK_BACK_SPACE)
			mozileDo("Delete backward");
			//Problem. Should doTyping if only a char deleted, but can't look ahead to see if structure is going to be deleted.
			//If structure deleted, could set undo key counter to maximum to force save next time?
			mozilePrefs.setBoolPref('accessibility.browsewithcaret', false);
			mozileSelection.deleteSelection();
		}else if(key==46){ // (DOM_VK_DELETE)
			mozileDo("Delete forward");
			mozilePrefs.setBoolPref('accessibility.browsewithcaret', false);
			mozileSelection.deleteSelection(mozileTarget.Selection.DELETE_FORWARD);
		}else{
			result=update=false;
		}
	}else if(C) { // normal character key. (NB event.isChar appears to be broken.)
		if(C==32 && shift)	//shift+spacebar -> non-break space.
			C=160;
		mozileDoTyping();
		mozilePrefs.setBoolPref('accessibility.browsewithcaret', false);
		mozileSelection.insertCharacter(C);
		mozile.source.browserInput(mozileTarget, mozileSelection);
		update=false;
	}else{
		result=update=false;
	}
	
	mozilePrefs.setBoolPref('accessibility.browsewithcaret', true);

	if(update){
		mozileDetect();
		mozile.source.browserInput(mozileTarget, mozileSelection);
	}
	return result;
}
